"""A CLI for the evohome-sync2 library."""

import sys

try:
    from evohomeasync2.client import main

except ModuleNotFoundError:
    import os

    sys.path.append(f"{os.path.dirname(__file__)}/src")

    from evohomeasync2.client import main

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 provides an async client for the *updated* Evohome API."""

import asyncio
import json
import logging
import os
import sys
from datetime import datetime as dt
from io import TextIOWrapper

import click

from . import HotWater, Zone
from .base import EvohomeClient
from .controlsystem import SZ_NAME, SZ_SCHEDULE, ControlSystem
from .schema.account import (
    SZ_ACCESS_TOKEN,
    SZ_ACCESS_TOKEN_EXPIRES,
    SZ_REFRESH_TOKEN,
)

# debug flags should be False for end-users
_DEBUG_CLI = False  # for debugging of CLI (*before* loading library)

DEBUG_ADDR = "0.0.0.0"
DEBUG_PORT = 5679

SZ_CACHE_TOKENS = "cache_tokens"
SZ_EVO = "evo"
SZ_USERNAME = "username"

TOKEN_FILE = ".evo-cache.tmp"


_LOGGER = logging.getLogger(__name__)


def _start_debugging(wait_for_client: bool) -> None:
    import debugpy  # type: ignore[import-untyped]

    debugpy.listen(address=(DEBUG_ADDR, DEBUG_PORT))
    print(f" - Debugging is enabled, listening on: {DEBUG_ADDR}:{DEBUG_PORT}")

    if wait_for_client:
        print("   - execution paused, waiting for debugger to attach...")
        debugpy.wait_for_client()
        print("   - debugger is now attached, continuing execution.")


if _DEBUG_CLI:
    _start_debugging(True)


def _check_zone_id(ctx: click.Context, param: click.Option, value: str) -> str:
    """Validate the zone_idx argument is "00" to "11", or "HW"."""

    return value

    # if value.upper() == "HW":
    #     return "HW"

    # if not value.isdigit() or int(value) not in range(0, 12):
    #     raise click.BadParameter("must be '00' to '11', or 'HW'")

    # return f"{int(value):02X}"


def _check_positive_int(ctx: click.Context, param: click.Option, value: int) -> int:
    """Validate the parameter is a positive int."""

    if value < 0:
        raise click.BadParameter("must >= 0")

    return value


def _get_tcs(evo: EvohomeClient, loc_idx: int | None) -> ControlSystem:
    """Get the ControlSystem object for the specified location idx."""

    if loc_idx is None:
        return evo._get_single_tcs()

    return evo.locations[int(loc_idx)]._gateways[0]._control_systems[0]


def _dump_tokens(evo: EvohomeClient) -> None:
    """Dump the tokens to a cache (temporary file)."""

    if evo.access_token_expires:
        expires = evo.access_token_expires.isoformat()
    else:
        expires = None

    with open(TOKEN_FILE, "w") as fp:
        json.dump(
            {
                # SZ_USERNAME: evo.username,
                SZ_REFRESH_TOKEN: evo.refresh_token,
                SZ_ACCESS_TOKEN: evo.access_token,
                SZ_ACCESS_TOKEN_EXPIRES: expires,
            },
            fp,
        )

    _LOGGER.warning("Access tokens cached to: %s", TOKEN_FILE)


def _load_tokens() -> dict[str, dt | str]:
    """Load the tokens from a cache (temporary file)."""

    if not os.path.exists(TOKEN_FILE):
        return {}

    with open(TOKEN_FILE) as f:
        tokens = json.load(f)

    if SZ_ACCESS_TOKEN_EXPIRES not in tokens:
        return tokens  # type: ignore[no-any-return]

    if expires := tokens[SZ_ACCESS_TOKEN_EXPIRES]:
        tokens[SZ_ACCESS_TOKEN_EXPIRES] = dt.fromisoformat(expires)

    return tokens  # type: ignore[no-any-return]


@click.group()
@click.option("--username", "-u", required=True, help="The TCC account username.")
@click.option("--password", "-p", required=True, help="The TCC account password.")
@click.option("--cache-tokens", "-c", is_flag=True, help="Cache of/for access tokens.")
@click.option("--debug", "-d", is_flag=True, help="Enable debug logging.")
@click.pass_context
def cli(
    ctx: click.Context,
    username: str,
    password: str,
    location: int | None = None,
    cache_tokens: bool | None = None,
    debug: bool | None = None,
) -> None:
    # if debug:  # Do first
    #     _start_debugging(True)

    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARNING,
        format="%(asctime)s %(levelname)s %(name)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )

    ctx.obj = ctx.obj or {}  # may be None
    ctx.obj[SZ_CACHE_TOKENS] = cache_tokens

    tokens = _load_tokens() if cache_tokens else {}

    ctx.obj[SZ_EVO] = EvohomeClient(
        username,
        password,
        debug=bool(debug),
        **tokens,  # type: ignore[arg-type]
    )


@cli.command()
@click.argument("zone_id", callback=_check_zone_id, type=str)
@click.option(  # --loc-idx
    "--loc-idx",
    "-l",
    callback=_check_positive_int,
    default=0,
    type=int,
    help="The location idx.",
)
@click.option(  # --filename
    "--filename", "-f", type=click.File("w"), default="-", help="The output file."
)
@click.pass_context
def get_schedule(
    ctx: click.Context, zone_id: str, loc_idx: int, filename: TextIOWrapper
) -> None:
    """Download the schedule of a zone of a TCS."""

    async def get_schedule(
        evo: EvohomeClient, zone_id: str, loc_idx: int | None
    ) -> None:
        try:
            await evo.login()

            tcs: ControlSystem = _get_tcs(evo, loc_idx)
            child: HotWater | Zone = tcs.zones_by_id[zone_id]

            schedule = await child.get_schedule()

        finally:  # FIXME: EvohomeClient should do this...
            assert evo.broker._session is not None  # mypy hint
            await evo.broker._session.close()  # FIXME

        schedules = {
            child._id: {SZ_NAME: child.name, SZ_SCHEDULE: schedule},
        }

        filename.write(json.dumps(schedules, indent=4))
        filename.write("\r\n\r\n")

    print("\r\nclient.py: Starting backup...")
    evo = ctx.obj[SZ_EVO]

    asyncio.run(get_schedule(evo, zone_id, loc_idx))

    if ctx.obj[SZ_CACHE_TOKENS]:
        _dump_tokens(evo)
    print(" - finished.\r\n")


@cli.command()
@click.option(  # --loc-idx
    "--loc-idx",
    "-l",
    callback=_check_positive_int,
    default=0,
    type=int,
    help="The location idx.",
)
@click.option(  # --filename
    "--filename", "-f", type=click.File("w"), default="-", help="The output file."
)
@click.pass_context
def get_schedules(ctx: click.Context, loc_idx: int, filename: TextIOWrapper) -> None:
    """Download all the schedule from a TCS."""

    async def get_schedules(evo: EvohomeClient, loc_idx: int | None) -> None:
        try:
            await evo.login()

            tcs: ControlSystem = _get_tcs(evo, loc_idx)
            schedules = await tcs._get_schedules()

        finally:  # FIXME: EvohomeClient should do this...
            assert evo.broker._session is not None  # mypy hint
            await evo.broker._session.close()  # FIXME

        filename.write(json.dumps(schedules, indent=4))
        filename.write("\r\n\r\n")

    print("\r\nclient.py: Starting backup...")
    evo = ctx.obj[SZ_EVO]

    asyncio.run(get_schedules(evo, loc_idx))

    if ctx.obj[SZ_CACHE_TOKENS]:
        _dump_tokens(evo)
    print(" - finished.\r\n")


@cli.command()
@click.option(  # --loc-idx
    "--loc-idx",
    "-l",
    callback=_check_positive_int,
    default=0,
    type=int,
    help="The location idx.",
)
@click.option("--filename", "-f", type=click.File(), help="The input file.")
@click.pass_context
def set_schedules(ctx: click.Context, loc_idx: int, filename: TextIOWrapper) -> None:
    """Upload schedules to a TCS."""

    async def set_schedules(evo: EvohomeClient, loc_idx: int | None) -> bool:
        schedules = json.loads(filename.read())

        try:
            await evo.login()

            tcs: ControlSystem = _get_tcs(evo, loc_idx)
            success = await tcs._set_schedules(schedules)

        finally:  # FIXME: EvohomeClient should do this...
            assert evo.broker._session is not None  # mypy hint
            await evo.broker._session.close()  # FIXME

        return success

    print("\r\nclient.py: Starting restore...")
    evo = ctx.obj[SZ_EVO]

    asyncio.run(set_schedules(evo, loc_idx))

    if ctx.obj[SZ_CACHE_TOKENS]:
        _dump_tokens(evo)
    print(" - finished.\r\n")


def main() -> None:
    try:
        cli(obj={})  # default for ctx.obj is None

    except click.ClickException as exc:
        print(f"Error: {exc}")
        sys.exit(-1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync provides an async client for the *original* Evohome API."""

from typing import Any, TypeAlias

# TCC config, status dicts
_EvoLeafT: TypeAlias = bool | float | int | str | list[str]  # Any
_DeviceDictT: TypeAlias = dict[str, Any]  # '_EvoDeviceT' | _EvoLeafT]
_EvoDictT: TypeAlias = dict[str, Any]  # '_EvoDictT' | _EvoLeafT]
_EvoListT: TypeAlias = list[_EvoDictT]
_EvoSchemaT: TypeAlias = _EvoDictT | _EvoListT

# TCC identifiers (Usr, Loc, Gwy, Sys, Zon|Dhw)
_DhwIdT: TypeAlias = int
_GatewayIdT: TypeAlias = int
_LocationIdT: TypeAlias = int
_SystemIdT: TypeAlias = int
_UserIdT: TypeAlias = int
_ZoneIdT: TypeAlias = int
_ZoneNameT: TypeAlias = str

# TCC other
_ModeT: TypeAlias = str
_SystemModeT: TypeAlias = str

_TaskIdT: TypeAlias = str  # TODO: int or str?


SZ_SESSION_ID = "sessionId"  # id Id, not ID

# schema keys (start with a lower case letter)
SZ_ALLOWED_MODES = "allowedModes"
SZ_CHANGEABLE_VALUES = "changeableValues"
SZ_COOL_SETPOINT = "coolSetpoint"
SZ_DEVICE_ID = "deviceID"  # is ID, not Id
SZ_DEVICES = "devices"
SZ_DOMAIN_ID = "domainID"  # is ID, not Id
SZ_GATEWAY_ID = "gatewayId"  # is Id, not ID
SZ_HEAT_SETPOINT = "heatSetpoint"
SZ_ID = "id"  # is id, not Id/ID
SZ_INDOOR_TEMPERATURE = "indoorTemperature"
SZ_LOCATION_ID = "locationID"  # is ID, not Id
SZ_MAC_ID = "macID"  # is ID, not Id
SZ_MODE = "mode"
SZ_NAME = "name"
SZ_NEXT_TIME = "NextTime"
SZ_QUICK_ACTION = "QuickAction"
SZ_QUICK_ACTION_NEXT_TIME = "QuickActionNextTime"
SZ_SETPOINT = "setpoint"
SZ_SPECIAL_MODES = "SpecialModes"
SZ_STATE = "state"
SZ_STATUS = "status"
SZ_TEMP = "temp"
SZ_THERMOSTAT = "thermostat"
SZ_THERMOSTAT_MODEL_TYPE = "thermostatModelType"
SZ_USER_ID = "userID"  # is ID, not Id
SZ_USER_INFO = "userInfo"
SZ_VALUE = "value"

# schema values (start with an upper case letter)
SZ_AUTO = "Auto"
SZ_AUTO_WITH_ECO = "AutoWithEco"
SZ_AWAY = "Away"
SZ_CUSTOM = "Custom"
SZ_DAY_OFF = "DayOff"
SZ_HEATING_OFF = "HeatingOff"
#
SZ_DHW_OFF = "DHWOff"
SZ_DHW_ON = "DHWOn"
#
SZ_DOMESTIC_HOT_WATER = "DOMESTIC_HOT_WATER"
SZ_EMEA_ZONE = "EMEA_ZONE"
#
SZ_HOLD = "Hold"
SZ_SCHEDULED = "Scheduled"
SZ_TEMPORARY = "Temporary"
#
SZ_HEAT = "Heat"
SZ_OFF = "Off"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 provides an async client for the *updated* Evohome API."""
from __future__ import annotations

import logging
from datetime import datetime as dt
from http import HTTPStatus
from typing import TYPE_CHECKING, NoReturn

import aiohttp

from . import exceptions as exc
from .broker import Broker
from .controlsystem import ControlSystem
from .location import Location
from .schema import SCH_FULL_CONFIG, SCH_USER_ACCOUNT

if TYPE_CHECKING:
    from .schema import (
        _EvoDictT,
        _EvoListT,
        _EvoSchemaT,
        _FilePathT,
        _LocationIdT,
        _SystemIdT,
    )


_LOGGER = logging.getLogger(__name__.rpartition(".")[0])


class EvohomeClientDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    async def full_installation(
        self, location_id: None | _LocationIdT = None
    ) -> NoReturn:
        # if location_id is None:
        #     location_id = self.installation_info[0]["locationInfo"]["locationId"]
        # url = f"location/{location_id}/installationInfo?"  # Specific location

        raise exc.DeprecationError(
            "EvohomeClient.full_installation() is deprecated, use .installation()"
        )

    async def gateway(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError("EvohomeClient.gateway() is deprecated")

    async def set_status_away(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_away() is deprecated, use .set_mode_away()"
        )

    async def set_status_custom(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_custom() is deprecated, use .set_mode_custom()"
        )

    async def set_status_dayoff(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_dayoff() is deprecated, use .set_mode_dayoff()"
        )

    async def set_status_eco(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_eco() is deprecated, use .set_mode_eco()"
        )

    async def set_status_heatingoff(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_heatingoff() is deprecated, use .set_mode_heatingoff()"
        )

    async def set_status_normal(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_normal() is deprecated, use .set_mode_auto()"
        )

    async def set_status_reset(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_reset() is deprecated, use .reset_mode()"
        )

    async def zone_schedules_backup(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.zone_schedules_backup() is deprecated, use .backup_schedules()"
        )

    async def zone_schedules_restore(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.zone_schedules_restore() is deprecated, use .restore_schedules()"
        )


class EvohomeClient(EvohomeClientDeprecated):
    """Provide a client to access the Honeywell TCC API."""

    _full_config: _EvoListT = None  # type: ignore[assignment]  # installation_info (all locations of user)
    _user_account: _EvoDictT = None  # type: ignore[assignment]  # account_info

    def __init__(
        self,
        username: str,
        password: str,
        /,
        *,
        refresh_token: str | None = None,
        access_token: str | None = None,
        access_token_expires: dt | None = None,
        session: None | aiohttp.ClientSession = None,
        debug: bool = False,
    ) -> None:
        """Construct the v2 EvohomeClient object.

        If access/refresh tokens are provided they will be used to avoid calling the
        authentication service, which is known to be rate limited.
        """

        if debug:
            _LOGGER.setLevel(logging.DEBUG)
            _LOGGER.debug("Debug mode is explicitly enabled.")

        self._logger = _LOGGER

        self.broker = Broker(
            username,
            password,
            _LOGGER,
            refresh_token=refresh_token,
            access_token=access_token,
            access_token_expires=access_token_expires,
            session=session,
        )

        self.locations: list[Location] = []

    @property
    def username(self) -> str:  # TODO: deprecate? or use config JSON?
        return self.broker._credentials["Username"]

    @property
    def password(self) -> str:  # TODO: deprecate
        return self.broker._credentials["Password"]

    @property
    def refresh_token(self) -> str | None:  # TODO: deprecate
        return self.broker.refresh_token

    @property
    def access_token(self) -> str | None:  # TODO: deprecate
        return self.broker.access_token

    @property
    def access_token_expires(self) -> dt | None:  # TODO: deprecate
        return self.broker.access_token_expires

    async def login(self) -> None:
        """Retrieve the user account and installation details.

        Will authenticate as required.
        """

        try:  # the cached access_token may be valid, but is not authorized
            await self.user_account()

        except exc.AuthenticationFailed as err:
            if err.status != HTTPStatus.UNAUTHORIZED or not self.access_token:
                raise

            _LOGGER.warning("Unauthorized access_token (will try re-authenticating).")
            self.broker.access_token = None  # FIXME: this is a hack
            await self.user_account(force_update=True)

        await self.installation()

    @property  # user_account
    def account_info(self) -> _EvoSchemaT:  # from original evohomeclient namespace
        """Return the information of the user account."""
        return self._user_account

    async def user_account(self, force_update: bool = False) -> _EvoDictT:
        """Return the user account information.

        If required/forced, retrieve that data from the vendor's API.
        """

        # There is usually no need to refresh this data (it is config, not state)
        if self._user_account and not force_update:
            return self._user_account

        # this code should really be in broker somewhere
        if not self.broker._session:
            self.broker._session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=30)
            )

        self._user_account = await self.broker.get(
            "userAccount", schema=SCH_USER_ACCOUNT
        )  # type: ignore[assignment]

        return self._user_account

    @property  # full_config (all locations of user)
    def installation_info(self) -> _EvoListT:  # from original evohomeclient namespace
        """Return the installation info (config) of all the user's locations."""
        return self._full_config

    async def installation(self, force_update: bool = False) -> _EvoListT:
        """Return the configuration of the user's locations their status.

        If required/forced, retrieve that data from the vendor's API.
        Note that the force_update flag will create new location entities (it includes
        `self.locations = []`).
        """

        # There is usually no need to refresh this data (it is config, not state)
        if self._full_config and not force_update:
            return self._full_config

        return await self._installation()  # aka self.installation_info

    async def _installation(self, refresh_status: bool = True) -> _EvoListT:
        """Return the configuration of the user's locations their status.

        The refresh_status flag is used for dev/test to disable retreiving the initial
        status of each location (and its child entities).
        """

        assert isinstance(self.account_info, dict)  # mypy

        # FIXME: shouldn't really be starting again with new objects?
        self.locations = []  # for now, need to clear this before GET

        url = f"location/installationInfo?userId={self.account_info['userId']}"
        url += "&includeTemperatureControlSystems=True"

        self._full_config = await self.broker.get(url, schema=SCH_FULL_CONFIG)  # type: ignore[assignment]

        # populate each freshly instantiated location with its initial status
        loc_config: _EvoDictT

        for loc_config in self._full_config:
            loc = Location(self, loc_config)
            self.locations.append(loc)
            if refresh_status:
                await loc.refresh_status()

        return self._full_config

    def _get_single_tcs(self) -> ControlSystem:
        """If there is a single location/gateway/TCS, return it, or raise an exception.

        Most users will have only one TCS.
        """

        if not self.locations or len(self.locations) != 1:
            raise exc.NoSingleTcsError(
                "There is not a single location (only) for this account"
            )

        if len(self.locations[0]._gateways) != 1:
            raise exc.NoSingleTcsError(
                "There is not a single gateway (only) for this account/location"
            )

        if len(self.locations[0]._gateways[0]._control_systems) != 1:
            raise exc.NoSingleTcsError(
                "There is not a single TCS (only) for this account/location/gateway"
            )

        return self.locations[0]._gateways[0]._control_systems[0]

    @property
    def system_id(self) -> _SystemIdT:  # an evohome-client anachronism, deprecate?
        """Return the ID of the default TCS (assumes only one loc/gwy/TCS)."""
        return self._get_single_tcs().systemId

    async def reset_mode(self) -> None:
        """Reset the mode of the default TCS and its zones."""
        await self._get_single_tcs().reset_mode()

    async def set_mode_auto(self) -> None:
        """Set the default TCS into auto mode."""
        await self._get_single_tcs().set_auto()

    async def set_mode_away(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into away mode."""
        await self._get_single_tcs().set_away(until=until)

    async def set_mode_custom(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into custom mode."""
        await self._get_single_tcs().set_custom(until=until)

    async def set_mode_dayoff(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into day off mode."""
        await self._get_single_tcs().set_dayoff(until=until)

    async def set_mode_eco(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into eco mode."""
        await self._get_single_tcs().set_eco(until=until)

    async def set_mode_heatingoff(self, /, *, until: dt | None = None) -> None:
        """Set the default TCS into heating off mode."""
        await self._get_single_tcs().set_heatingoff(until=until)

    async def temperatures(self) -> _EvoListT:
        """Return the current temperatures and setpoints of the default TCS."""
        return await self._get_single_tcs().temperatures()

    async def backup_schedules(self, filename: _FilePathT) -> None:
        """Backup all schedules from the default TCS to the file."""
        await self._get_single_tcs().backup_schedules(filename)

    async def restore_schedules(
        self, filename: _FilePathT, match_by_name: bool = False
    ) -> None:
        """Restore all schedules from the file to the control system.

        There is the option to match schedules to their zone/dhw by name rather than id.
        """

        await self._get_single_tcs().restore_schedules(
            filename, match_by_name=match_by_name
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the config & status schemas."""
from __future__ import annotations

import json
from pathlib import Path

import pytest

from evohomeasync2.schema import vol

TEST_DIR = Path(__file__).resolve().parent


def _test_schema(folder: Path, schema: vol.Schema, file_name: str):  # type: ignore[no-any-unimported]
    if not Path(folder).joinpath(file_name).is_file():
        pytest.skip(f"No {file_name} in: {folder.name}")

    with open(Path(folder).joinpath(file_name)) as f:
        data: dict = json.load(f)

    _ = schema(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync provides an async client for the *original* Evohome API."""
from __future__ import annotations

import asyncio
import logging
from datetime import datetime as dt
from http import HTTPMethod
from typing import TYPE_CHECKING, NoReturn

import aiohttp

from . import exceptions as exc
from .broker import Broker, _LocnDataT, _SessionIdT, _UserDataT, _UserInfoT
from .schema import (
    SZ_ALLOWED_MODES,
    SZ_AUTO,
    SZ_AUTO_WITH_ECO,
    SZ_AWAY,
    SZ_CHANGEABLE_VALUES,
    SZ_CUSTOM,
    SZ_DAY_OFF,
    SZ_DEVICE_ID,
    SZ_DEVICES,
    SZ_DHW_OFF,
    SZ_DHW_ON,
    SZ_DOMESTIC_HOT_WATER,
    SZ_EMEA_ZONE,
    SZ_HEAT_SETPOINT,
    SZ_HEATING_OFF,
    SZ_HOLD,
    SZ_ID,
    SZ_INDOOR_TEMPERATURE,
    SZ_LOCATION_ID,
    SZ_MODE,
    SZ_NAME,
    SZ_NEXT_TIME,
    SZ_QUICK_ACTION,
    SZ_QUICK_ACTION_NEXT_TIME,
    SZ_SCHEDULED,
    SZ_SESSION_ID,
    SZ_SETPOINT,
    SZ_STATE,
    SZ_STATUS,
    SZ_TEMP,
    SZ_TEMPORARY,
    SZ_THERMOSTAT,
    SZ_THERMOSTAT_MODEL_TYPE,
    SZ_USER_INFO,
    SZ_VALUE,
)

if TYPE_CHECKING:
    from .schema import (
        _DeviceDictT,
        _DhwIdT,
        _EvoListT,
        _LocationIdT,
        _SystemModeT,
        _TaskIdT,
        _ZoneIdT,
        _ZoneNameT,
    )


_LOGGER = logging.getLogger(__name__.rpartition(".")[0])


class EvohomeClientDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    @property
    def user_data(self) -> _UserDataT | None:
        raise exc.DeprecationError(
            "EvohomeClient.user_data is deprecated, use .user_info"
            " (session_id is now .broker.session_id)"
        )

    @property
    def full_data(self) -> _UserDataT | None:
        raise exc.DeprecationError(
            "EvohomeClient.full_data is deprecated, use .location_data"
        )

    @property
    def headers(self) -> str:
        raise exc.DeprecationError("EvohomeClient.headers is deprecated")

    @property
    def hostname(self) -> str:
        raise exc.DeprecationError(
            "EvohomeClient.hostanme is deprecated, use .broker.hostname"
        )

    @property
    def postdata(self) -> str:
        raise exc.DeprecationError("EvohomeClient.postdata is deprecated")

    async def _wait_for_put_task(self, response: aiohttp.ClientResponse) -> None:
        """This functionality is deprecated, but remains here as documentation."""

        async def get_task_status(task_id: _TaskIdT) -> str:
            await self._populate_locn_data()

            url = f"/commTasks?commTaskId={task_id}"
            response = await self._do_request(HTTPMethod.GET, url)

            ret: str = dict(await response.json())[SZ_STATE]
            return ret

        task_id: _TaskIdT

        assert response.method == HTTPMethod.PUT

        ret = await response.json()
        task_id = ret[0][SZ_ID] if isinstance(ret, list) else ret[SZ_ID]

        # FIXME: could wait forvever?
        while await get_task_status(task_id) != "Succeeded":
            await asyncio.sleep(1)

    # Not deprecated, just a placeholder for self._wait_for_put_task()
    async def _do_request(self, *args, **kwargs) -> aiohttp.ClientResponse:  # type: ignore[no-untyped-def]
        raise NotImplementedError

    # Not deprecated, just a placeholder for self._wait_for_put_task()
    async def _populate_locn_data(self, force_refresh: bool = True) -> _LocnDataT:
        raise NotImplementedError

    async def get_system_modes(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.get_modes() is deprecated, "
            "use .get_system_modes() or .get_zone_modes()"
        )

    async def set_status_away(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_away() is deprecated, use .set_mode_away()"
        )

    async def set_status_custom(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_custom() is deprecated, use .set_mode_custom()"
        )

    async def set_status_dayoff(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_dayoff() is deprecated, use .set_mode_dayoff()"
        )

    async def set_status_eco(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_eco() is deprecated, use .set_mode_eco()"
        )

    async def set_status_heatingoff(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_heatingoff() is deprecated, use .set_mode_heatingoff()"
        )

    async def set_status_normal(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.set_status_normal() is deprecated, use .set_mode_auto()"
        )

    async def temperatures(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.temperatures() is deprecated, use .get_temperatures()"
        )

    async def cancel_temp_override(self, *args, **kwargs) -> NoReturn:  # type: ignore[no-untyped-def]
        raise exc.DeprecationError(
            "EvohomeClient.cancel_temp_override() is deprecated, use .set_zone_auto()"
        )


# any API request invokes self._populate_user_data()             (for authentication)
# - every API GET invokes self._populate_locn_data(refresh=True) (for up-to-date state)
# - every API PUT invokes self._populate_locn_data()             (for config)


class EvohomeClient(EvohomeClientDeprecated):
    """Provide a client to access the Honeywell TCC API (assumes a single TCS)."""

    _LOC_IDX: int = 0  # the index of the location in the full_data list

    user_info: _UserInfoT  # user_data[SZ_USER_INFO] only (i.e. *without* "sessionID")
    location_data: _LocnDataT  # of the first location (config and status) in list

    def __init__(
        self,
        username: str,
        password: str,
        /,
        *,
        session_id: _SessionIdT | None = None,
        session: aiohttp.ClientSession | None = None,
        hostname: str | None = None,  # is a URL
        debug: bool = False,
    ) -> None:
        """Construct the v1 EvohomeClient object.

        If a session_id is provided it will be used to avoid calling the
        authentication service, which is known to be rate limited.
        """
        if debug:
            _LOGGER.setLevel(logging.DEBUG)
            _LOGGER.debug("Debug mode is explicitly enabled.")

        self.user_info = {}
        self.location_data = {}
        self.location_id: _LocationIdT = None  # type: ignore[assignment]

        self.devices: dict[_ZoneIdT, _DeviceDictT] = {}  # dhw or zone by id
        self.named_devices: dict[_ZoneNameT, _DeviceDictT] = {}  # zone by name

        self.broker = Broker(
            username,
            password,
            _LOGGER,
            session_id=session_id,
            hostname=hostname,
            session=session,
        )

    @property
    def user_data(self) -> _UserDataT | None:  # TODO: deprecate?
        """Return the user data used for HTTP authentication."""

        if not self.broker.session_id:
            return None
        return {
            SZ_SESSION_ID: self.broker.session_id,
            SZ_USER_INFO: self.user_info,
        }

    # User methods...

    async def _populate_user_data(
        self, force_refresh: bool = False
    ) -> dict[str, bool | int | str]:
        """Retrieve the cached user data (excl. the session ID).

        Pull the latest JSON from the web only if force_refresh is True.
        """

        if not self.user_info or force_refresh:
            user_data = await self.broker.populate_user_data()
            self.user_info = user_data[SZ_USER_INFO]  # type: ignore[assignment]

        return self.user_info  # excludes session ID

    async def _get_user(self) -> _UserInfoT:
        """Return the user (if needed, get the JSON)."""

        # only retrieve the config data if we don't already have it
        if not self.user_info:
            await self._populate_user_data(force_refresh=False)
        return self.user_info

    # Location methods...

    async def _populate_locn_data(self, force_refresh: bool = True) -> _LocnDataT:
        """Retrieve the latest system data.

        Pull the latest JSON from the web unless force_refresh is False.
        """

        if not self.location_data or force_refresh:
            full_data = await self.broker.populate_full_data()
            self.location_data = full_data[self._LOC_IDX]

            self.location_id = self.location_data[SZ_LOCATION_ID]

            self.devices = {d[SZ_DEVICE_ID]: d for d in self.location_data[SZ_DEVICES]}
            self.named_devices = {d[SZ_NAME]: d for d in self.location_data[SZ_DEVICES]}

        return self.location_data

    async def get_temperatures(
        self, force_refresh: bool = True
    ) -> _EvoListT:  # a convenience function
        """Retrieve the latest details for each zone (incl. DHW)."""

        set_point: float
        status: str

        await self._populate_locn_data(force_refresh=force_refresh)

        result = []

        try:
            for device in self.location_data[SZ_DEVICES]:
                temp = float(device[SZ_THERMOSTAT][SZ_INDOOR_TEMPERATURE])
                values = device[SZ_THERMOSTAT][SZ_CHANGEABLE_VALUES]

                if SZ_HEAT_SETPOINT in values:
                    set_point = float(values[SZ_HEAT_SETPOINT][SZ_VALUE])
                    status = values[SZ_HEAT_SETPOINT][SZ_STATUS]
                else:
                    set_point = 0
                    status = values[SZ_STATUS]

                result.append(
                    {
                        SZ_THERMOSTAT: device[SZ_THERMOSTAT_MODEL_TYPE],
                        SZ_ID: device[SZ_DEVICE_ID],
                        SZ_NAME: device[SZ_NAME],
                        SZ_TEMP: None if temp == 128 else temp,
                        SZ_SETPOINT: set_point,
                        SZ_STATUS: status,
                        SZ_MODE: values[SZ_MODE],
                    }
                )

        # harden code against unexpected schema (JSON structure)
        except (LookupError, TypeError, ValueError) as err:
            raise exc.InvalidSchema(str(err)) from err
        return result

    async def get_system_modes(self) -> NoReturn:
        """Return the set of modes the system can be assigned."""
        raise NotImplementedError

    async def _set_system_mode(
        self, status: _SystemModeT, until: dt | None = None
    ) -> None:
        """Set the system mode."""

        # just want id, so retrieve the config data only if we don't already have it
        await self._populate_locn_data(force_refresh=False)  # get self.location_id

        data = {SZ_QUICK_ACTION: status}
        if until:
            data |= {SZ_QUICK_ACTION_NEXT_TIME: until.strftime("%Y-%m-%dT%H:%M:%SZ")}

        url = f"/evoTouchSystems?locationId={self.location_id}"
        await self.broker.make_request(HTTPMethod.PUT, url, data=data)

    async def set_mode_auto(self) -> None:
        """Set the system to normal operation."""
        await self._set_system_mode(SZ_AUTO)

    async def set_mode_away(self, until: dt | None = None) -> None:
        """Set the system to the away mode."""
        await self._set_system_mode(SZ_AWAY, until)

    async def set_mode_custom(self, until: dt | None = None) -> None:
        """Set the system to the custom programme."""
        await self._set_system_mode(SZ_CUSTOM, until)

    async def set_mode_dayoff(self, until: dt | None = None) -> None:
        """Set the system to the day off mode."""
        await self._set_system_mode(SZ_DAY_OFF, until)

    async def set_mode_eco(self, until: dt | None = None) -> None:
        """Set the system to the eco mode."""
        await self._set_system_mode(SZ_AUTO_WITH_ECO, until)

    async def set_mode_heatingoff(self, until: dt | None = None) -> None:
        """Set the system to the heating off mode."""
        await self._set_system_mode(SZ_HEATING_OFF, until)

    # Zone methods...

    async def _get_zone(self, id_or_name: _ZoneIdT | _ZoneNameT) -> _DeviceDictT:
        """Return the location's zone by its id or name (if needed, get the JSON).

        Raise an exception if the zone is not found.
        """

        device: _DeviceDictT

        # just want id, so retrieve the config data only if we don't already have it
        await self._populate_locn_data(force_refresh=False)

        if isinstance(id_or_name, int):
            device = self.devices.get(id_or_name)  # type: ignore[assignment]
        else:
            device = self.named_devices.get(id_or_name)  # type: ignore[assignment]

        if device is None:
            raise exc.InvalidSchema(
                f"No zone {id_or_name} in location {self.location_id}"
            )

        if (model := device[SZ_THERMOSTAT_MODEL_TYPE]) != SZ_EMEA_ZONE:
            raise exc.InvalidSchema(f"Zone {id_or_name} is not an EMEA_ZONE: {model}")

        assert device is not None  # mypy check

        return device

    async def get_zone_modes(self, zone: _ZoneNameT) -> list[str]:
        """Return the set of modes the zone can be assigned."""

        device = await self._get_zone(zone)

        ret: list[str] = device[SZ_THERMOSTAT][SZ_ALLOWED_MODES]
        return ret

    async def _set_heat_setpoint(
        self,
        zone: _ZoneIdT | _ZoneNameT,
        status: str,  # "Scheduled" | "Temporary" | "Hold
        value: float | None = None,
        next_time: dt | None = None,  # "%Y-%m-%dT%H:%M:%SZ"
    ) -> None:
        """Set zone setpoint, either indefinitely, or until a set time."""

        zone_id: _ZoneIdT = (await self._get_zone(zone))[SZ_DEVICE_ID]

        if next_time is None:
            data = {SZ_STATUS: SZ_HOLD, SZ_VALUE: value}
        else:
            data = {
                SZ_STATUS: status,
                SZ_VALUE: value,
                SZ_NEXT_TIME: next_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
            }

        url = f"/devices/{zone_id}/thermostat/changeableValues/heatSetpoint"
        await self.broker.make_request(HTTPMethod.PUT, url, data=data)

    async def set_temperature(
        self, zone: _ZoneIdT | _ZoneNameT, temperature: float, until: dt | None = None
    ) -> None:
        """Override the setpoint of a zone, for a period of time, or indefinitely."""

        if until:
            await self._set_heat_setpoint(
                zone, SZ_TEMPORARY, value=temperature, next_time=until
            )
        else:
            await self._set_heat_setpoint(zone, SZ_HOLD, value=temperature)

    async def set_zone_auto(self, zone: _ZoneIdT | _ZoneNameT) -> None:
        """Set a zone to follow its schedule."""
        await self._set_heat_setpoint(zone, status=SZ_SCHEDULED)

    # DHW methods...

    async def _get_dhw(self) -> _DeviceDictT:
        """Return the locations's DHW, if there is one (if needed, get the JSON).

        Raise an exception if the DHW is not found.
        """

        # just want id, so retrieve the config data only if we don't already have it
        await self._populate_locn_data(force_refresh=False)

        for device in self.location_data[SZ_DEVICES]:
            if device[SZ_THERMOSTAT_MODEL_TYPE] == SZ_DOMESTIC_HOT_WATER:
                ret: _DeviceDictT = device
                return ret

        raise exc.InvalidSchema(f"No DHW in location {self.location_id}")

    async def _set_dhw(
        self,
        status: str,  # "Scheduled" | "Hold"
        mode: str | None = None,  # "DHWOn" | "DHWOff"
        next_time: dt | None = None,  # "%Y-%m-%dT%H:%M:%SZ"
    ) -> None:
        """Set DHW to Auto, or On/Off, either indefinitely, or until a set time."""

        dhw_id: _DhwIdT = (await self._get_dhw())[SZ_DEVICE_ID]

        data = {
            SZ_STATUS: status,
            SZ_MODE: mode,
            # SZ_NEXT_TIME: None,
            # SZ_SPECIAL_TIMES: None,
            # SZ_HEAT_SETPOINT: None,
            # SZ_COOL_SETPOINT: None,
        }
        if next_time:
            data |= {SZ_NEXT_TIME: next_time.strftime("%Y-%m-%dT%H:%M:%SZ")}

        url = f"/devices/{dhw_id}/thermostat/changeableValues"
        await self.broker.make_request(HTTPMethod.PUT, url, data=data)

    async def set_dhw_on(self, until: dt | None = None) -> None:
        """Set DHW to On, either indefinitely, or until a specified time.

        When On, the DHW controller will work to keep its target temperature at/above
        its target temperature.  After the specified time, it will revert to its
        scheduled behaviour.
        """

        await self._set_dhw(status=SZ_HOLD, mode=SZ_DHW_ON, next_time=until)

    async def set_dhw_off(self, until: dt | None = None) -> None:
        """Set DHW to Off, either indefinitely, or until a specified time.

        When Off, the DHW controller will ignore its target temperature. After the
        specified time, it will revert to its scheduled behaviour.
        """

        await self._set_dhw(status=SZ_HOLD, mode=SZ_DHW_OFF, next_time=until)

    async def set_dhw_auto(self) -> None:
        """Allow DHW to switch between On and Off, according to its schedule."""
        await self._set_dhw(status=SZ_SCHEDULED)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the config & status schemas."""
from __future__ import annotations

import json
import logging
from pathlib import Path

import pytest

from evohomeasync2 import ControlSystem
from evohomeasync2.schema import SCH_LOCN_STATUS
from evohomeasync2.schema.config import SCH_TEMPERATURE_CONTROL_SYSTEM, SCH_TIME_ZONE
from evohomeasync2.schema.const import (
    SZ_GATEWAYS,
    SZ_LOCATION_INFO,
    SZ_TEMPERATURE_CONTROL_SYSTEMS,
    SZ_TIME_ZONE,
)

from .helpers import TEST_DIR

WORK_DIR = f"{TEST_DIR}/schemas"


class GatewayStub:
    def __init__(self) -> None:
        self._broker = None
        self._logger = logging.getLogger(__name__)
        self.location = None


def pytest_generate_tests(metafunc: pytest.Metafunc):
    def id_fnc(folder_path: Path):
        return folder_path.name

    folders = [
        p for p in Path(WORK_DIR).glob("*") if p.is_dir() and not p.name.startswith("_")
    ]
    metafunc.parametrize("folder", sorted(folders), ids=id_fnc)


# NOTE: JSON is not compliant with UserInfo schema, but is useful to test against TCS
def test_config_schemas(folder: Path):
    """Test the config schema for a location."""
    # Use pytest --log-cli-level=WARNING to see the output

    FILE_NAME = "config.json"

    if not Path(folder).joinpath(FILE_NAME).is_file():
        pytest.skip(f"No {FILE_NAME} in: {folder.name}")

    with open(Path(folder).joinpath(FILE_NAME)) as f:
        data: dict = json.load(f)

    _ = SCH_TIME_ZONE(data[SZ_LOCATION_INFO][SZ_TIME_ZONE])
    for gwy_config in data[SZ_GATEWAYS]:
        for tcs_config in gwy_config[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
            _ = SCH_TEMPERATURE_CONTROL_SYSTEM(tcs_config)
            _ = ControlSystem(GatewayStub(), tcs_config)


def test_status_schemas(folder: Path):
    """Test the status schema for a location."""

    FILE_NAME = "status.json"

    if not Path(folder).joinpath(FILE_NAME).is_file():
        pytest.skip(f"No {FILE_NAME} in: {folder.name}")

    with open(Path(folder).joinpath(FILE_NAME)) as f:
        data: dict = json.load(f)

    _ = SCH_LOCN_STATUS(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the config, status & schedule schemas."""
from __future__ import annotations

import asyncio
import logging
import os
from datetime import datetime as dt
from http import HTTPMethod, HTTPStatus

import evohomeasync2 as evohome
from evohomeasync2.const import URL_BASE
from evohomeasync2.schema import vol  # type: ignore[import-untyped]

from . import _DEBUG_DISABLE_STRICT_ASSERTS, _DEBUG_USE_REAL_AIOHTTP
from .mocked_server import MockedServer

if _DEBUG_USE_REAL_AIOHTTP:
    import aiohttp
else:
    from .mocked_server import aiohttp  # type: ignore[no-redef]


_LOGGER = logging.getLogger(__name__)


def user_credentials():
    username = os.getenv("PYTEST_USERNAME") or "spotty.blackcat@gmail.com"  # secret
    password = os.getenv("PYTEST_PASSWORD") or "ziQajn732m5JYQ!"  # .         password

    return username, password


def client_session():
    if _DEBUG_USE_REAL_AIOHTTP:
        return aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30))
    return aiohttp.ClientSession(mocked_server=MockedServer(None, None))


def extract_oauth_tokens(evo: evohome.EvohomeClient):
    return (
        evo.refresh_token,
        evo.access_token,
        evo.access_token_expires,
    )


async def should_work(  # type: ignore[no-any-unimported]
    evo: evohome.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    schema: vol.Schema | None = None,
) -> dict | list:
    """Make a request that is expected to succeed."""

    response: aiohttp.ClientResponse

    response, content = await evo.broker._client(method, f"{URL_BASE}/{url}", json=json)

    response.raise_for_status()

    assert response.content_type == content_type

    if schema:
        return schema(content)
    return content


async def should_fail(
    evo: evohome.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    status: HTTPStatus | None = None,
) -> aiohttp.ClientResponse:
    """Make a request that is expected to fail."""

    response: aiohttp.ClientResponse

    response, content = await evo.broker._client(method, f"{URL_BASE}/{url}", json=json)

    try:
        response.raise_for_status()
    except aiohttp.ClientResponseError as exc:
        assert exc.status == status
    else:
        assert False, response.status

    if _DEBUG_DISABLE_STRICT_ASSERTS:
        return response

    assert response.content_type == content_type

    if isinstance(content, list):
        assert status in (
            HTTPStatus.BAD_REQUEST,
            HTTPStatus.NOT_FOUND,  # CommTaskNotFound
            HTTPStatus.UNAUTHORIZED,
        ), content
        assert "message" in content[0]  # sometimes "code" too

    elif isinstance(content, dict):
        assert status in (
            HTTPStatus.NOT_FOUND,
            HTTPStatus.METHOD_NOT_ALLOWED,
        ), content
        assert "message" in content  # sometimes "code" too

    elif isinstance(content, str):  # 404
        assert status in (HTTPStatus.NOT_FOUND,), content

    else:
        assert False, response.status

    return response


async def wait_for_comm_task(
    evo: evohome.EvohomeClient,
    task_id: str,
    timeout: int = 3,
) -> bool | None:
    """Wait for a communication task (API call) to complete."""

    await asyncio.sleep(0.5)

    url = f"commTasks?commTaskId={task_id}"

    start_time = dt.now()
    while True:
        response = await should_work(evo, HTTPMethod.GET, url)
        if response["state"] == "Succeeded":  # type: ignore[call-overload]
            return True
        if (dt.now() - start_time).total_seconds() > timeout:
            return False
        if response["state"] in ("Created", "Running"):  # type: ignore[call-overload]
            await asyncio.sleep(0.3)
            continue
        else:
            # raise RuntimeError(f"Unexpected state: {response['state']}")
            _LOGGER.warning(f"Unexpected state: {response['state']}")  # type: ignore[call-overload]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the handling of vendor APIs (URLs)."""
from __future__ import annotations

import logging
from http import HTTPMethod, HTTPStatus

import pytest
import pytest_asyncio

import evohomeasync as evohome
from evohomeasync.broker import URL_HOST

# FIXME: need v1 schemas
from evohomeasync2.schema import vol  # type: ignore[import-untyped]

from . import _DEBUG_DISABLE_STRICT_ASSERTS, _DEBUG_USE_REAL_AIOHTTP
from .helpers import (
    aiohttp,  # aiohttp may be mocked
    client_session as _client_session,
    user_credentials as _user_credentials,
)

URL_BASE = f"{URL_HOST}/WebAPI/api"


_LOGGER = logging.getLogger(__name__)

_global_user_data: str = None  # session_id


@pytest.fixture(autouse=True)
def patches_for_tests(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("evohomeasync.base.aiohttp", aiohttp)
    monkeypatch.setattr("evohomeasync.broker.aiohttp", aiohttp)


@pytest.fixture()
def user_credentials():
    return _user_credentials()


@pytest_asyncio.fixture
async def session():
    client_session = _client_session()
    try:
        yield client_session
    finally:
        await client_session.close()


async def instantiate_client(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
):
    """Instantiate a client, and logon to the vendor API."""

    global _global_user_data

    # Instantiation, NOTE: No API calls invoked during instantiation
    evo = evohome.EvohomeClient(
        username,
        password,
        session=session,
        session_id=_global_user_data,
    )

    # Authentication
    await evo._populate_user_data()
    _global_user_data = evo.broker.session_id

    return evo


async def should_work(  # type: ignore[no-any-unimported]
    evo: evohome.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    schema: vol.Schema | None = None,
) -> dict | list:
    """Make a request that is expected to succeed."""

    response: aiohttp.ClientResponse

    response = await evo._do_request(method, f"{URL_BASE}/{url}", data=json)

    response.raise_for_status()

    assert True or response.content_type == content_type

    if response.content_type == "application/json":
        content = await response.json()
    else:
        content = await response.text()

    if schema:
        return schema(content)
    return content


async def should_fail(
    evo: evohome.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    status: HTTPStatus | None = None,
) -> aiohttp.ClientResponse:
    """Make a request that is expected to fail."""

    response: aiohttp.ClientResponse

    response = await evo._do_request(method, f"{URL_BASE}/{url}", data=json)

    try:
        response.raise_for_status()
    except aiohttp.ClientResponseError as exc:
        assert exc.status == status
    else:
        assert False, response.status

    if _DEBUG_DISABLE_STRICT_ASSERTS:
        return response

    assert True or response.content_type == content_type

    if response.content_type == "application/json":
        content = await response.json()
    else:
        content = await response.text()

    if isinstance(content, dict):
        assert True or "message" in content
    elif isinstance(content, list):
        assert True or "message" in content[0]

    return content


async def _test_client_apis(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
):
    """Instantiate a client, and logon to the vendor API."""

    global _global_user_data

    # Instantiation, NOTE: No API calls invoked during instantiation
    evo = evohome.EvohomeClient(
        username,
        password,
        session=session,
        session_id=_global_user_data,
    )

    user_data = await evo._populate_user_data()
    assert user_data == evo.user_info

    _global_user_data = evo.broker.session_id

    full_data = await evo._populate_locn_data()
    assert full_data == evo.location_data

    temps = await evo.get_temperatures()

    assert temps

    # for _ in range(3):
    #     await asyncio.sleep(5)
    #     _ = await evo.get_temperatures()
    #     _LOGGER.warning("get_temperatures() OK")


@pytest.mark.asyncio
async def test_client_apis(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test _populate_user_data() & _populate_full_data()"""

    if not _DEBUG_USE_REAL_AIOHTTP:
        pytest.skip("Mocked server not implemented")

    try:
        await _test_client_apis(*user_credentials, session=session)
    except evohome.AuthenticationFailed:
        pytest.skip("Unable to authenticate")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the helper functions."""
from __future__ import annotations

import json
from pathlib import Path

# import pytest
from evohomeasync2.schema.schedule import (
    SCH_GET_SCHEDULE_ZONE,
    SCH_PUT_SCHEDULE_ZONE,
    convert_to_get_schedule,
    convert_to_put_schedule,
)

from .helpers import TEST_DIR

WORK_DIR = Path(f"{TEST_DIR}/schedules")


# def pytest_generate_tests(metafunc: pytest.Metafunc):
#     def id_fnc(folder_path: Path):
#         return folder_path.name

#     files = [
#         f for f in Path(WORK_DIR).glob("*.json") if not f.name.startswith("_")
#     ]
#     metafunc.parametrize("file", sorted(files), ids=id_fnc)


def test_schedule_to_pascal_case() -> None:
    """Convert a schedule to snake_case, and back again."""

    get_schedule = json.load(
        Path(f"{TEST_DIR}/systems/system_001/schedule_zone.json").open()
    )
    assert get_schedule == SCH_GET_SCHEDULE_ZONE(get_schedule)

    put_schedule = convert_to_put_schedule(get_schedule)
    assert put_schedule == SCH_PUT_SCHEDULE_ZONE(put_schedule)

    assert get_schedule == convert_to_get_schedule(put_schedule)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync provides an async client for the *original* Evohome API."""
from __future__ import annotations

import logging
from collections.abc import Callable
from datetime import datetime as dt
from http import HTTPMethod, HTTPStatus
from typing import Any, Final, TypeAlias

import aiohttp

from . import exceptions as exc
from .schema import (
    SZ_SESSION_ID,
    SZ_USER_ID,
    SZ_USER_INFO,
)

_SessionIdT: TypeAlias = str
_UserIdT: TypeAlias = int

_UserInfoT: TypeAlias = dict[str, bool | int | str]
_UserDataT: TypeAlias = dict[str, _SessionIdT | _UserInfoT]
_LocnDataT: TypeAlias = dict[str, Any]


URL_HOST = "https://tccna.honeywell.com"
_APP_ID = "91db1612-73fd-4500-91b2-e63b069b185c"

_LOGGER = logging.getLogger(__name__)


class Broker:
    """Provide a client to access the Honeywell TCC API (assumes a single TCS)."""

    _user_data: _UserDataT
    _full_data: list[_LocnDataT]

    def __init__(
        self,
        username: str,
        password: str,
        logger: logging.Logger,
        /,
        *,
        session_id: _SessionIdT | None = None,
        hostname: str | None = None,  # is a URL
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        """A class for interacting with the v1 Evohome API."""

        self.username = username
        self._logger = logger

        self._session_id: _SessionIdT | None = session_id
        self._user_id: _UserIdT | None = None

        self.hostname: Final[str] = hostname or URL_HOST
        self._session = session or aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=30)
        )

        self._headers: dict[str, str] = {
            "content-type": "application/json"
        }  # NB: no sessionId yet
        self._POST_DATA: Final[dict[str, str]] = {
            "Username": self.username,
            "Password": password,
            "ApplicationId": _APP_ID,
        }

        self._user_data = {}
        self._full_data = []

    @property
    def session_id(self) -> _SessionIdT | None:
        """Return the session id used for HTTP authentication."""
        return self._session_id

    async def populate_user_data(self) -> _UserDataT:
        """Return the latest user data as retrieved from the web."""

        user_data: _UserDataT

        user_data, _ = await self._populate_user_data()
        return user_data

    async def _populate_user_data(self) -> tuple[_UserDataT, aiohttp.ClientResponse]:
        """Return the latest user data as retrieved from the web."""

        url = "/session"
        response = await self.make_request(HTTPMethod.POST, url, data=self._POST_DATA)

        self._user_data: _UserDataT = await response.json()

        user_id: _UserIdT = self._user_data[SZ_USER_INFO][SZ_USER_ID]  # type: ignore[assignment,index]
        session_id: _SessionIdT = self._user_data[SZ_SESSION_ID]  # type: ignore[assignment]

        self._user_id = user_id
        self._session_id = self._headers[SZ_SESSION_ID] = session_id
        self._session_renewed = dt.now()

        self._logger.info(f"user_data = {self._user_data}")
        return self._user_data, response

    async def populate_full_data(self) -> list[_LocnDataT]:
        """Return the latest location data exactly as retrieved from the web."""

        if not self._user_id:  # not yet authenticated
            # maybe was instantiated with a bad session_id, so must check user_id
            await self.populate_user_data()

        url = f"/locations?userId={self._user_id}&allData=True"
        response = await self.make_request(HTTPMethod.GET, url, data=self._POST_DATA)

        self._full_data: list[_LocnDataT] = await response.json()

        self._logger.info(f"full_data = {self._full_data}\r\n")
        return self._full_data

    async def _make_request(
        self,
        func: Callable,  # type: ignore[type-arg]
        url: str,
        /,
        *,
        data: dict[str, Any] | None = None,
        _dont_reauthenticate: bool = False,  # used only with recursive call
    ) -> aiohttp.ClientResponse:
        """Perform an HTTP request, with an optional retry if re-authenticated."""

        response: aiohttp.ClientResponse

        async with func(url, json=data, headers=self._headers) as response:  # NB: json=
            response_text = await response.text()  # why cant I move this below the if?

            # if 401/unauthorized, may need to refresh sessionId (expires in 15 mins?)
            if response.status != HTTPStatus.UNAUTHORIZED or _dont_reauthenticate:
                return response

            # TODO: use response.content_type to determine whether to use .json()
            if "code" not in response_text:  # don't use .json() yet: may be plain text
                return response

            response_json = await response.json()
            if response_json[0]["code"] != "Unauthorized":
                return response

            # if SZ_SESSION_ID not in self._headers:  # no value trying to re-authenticate
            #     return response  # ...because: the user credentials must be invalid

            _LOGGER.debug(f"Session now expired/invalid ({self._session_id})...")
            self._headers = {"content-type": "application/json"}  # remove the sessionId

            _, response = await self._populate_user_data()  # Get a fresh sessionId
            assert self._session_id is not None  # mypy hint

            _LOGGER.debug(f"... success: new sessionId = {self._session_id}\r\n")
            self._headers[SZ_SESSION_ID] = self._session_id

            if "session" in url:  # retry not needed for /session
                return response

            # NOTE: this is a recursive call, used only after re-authenticating
            response = await self._make_request(
                func, url, data=data, _dont_reauthenticate=True
            )
            return response

    async def make_request(
        self,
        method: HTTPMethod,
        url: str,
        /,
        *,
        data: dict[str, Any] | None = None,
    ) -> aiohttp.ClientResponse:
        """Perform an HTTP request, will authenticate if required."""

        url = self.hostname + "/WebAPI/api" + url

        if method == HTTPMethod.GET:
            func = self._session.get
        elif method == HTTPMethod.PUT:
            func = self._session.put
        elif method == HTTPMethod.POST:
            func = self._session.post

        try:
            response = await self._make_request(func, url, data=data)

        except aiohttp.ClientError as err:
            if method == HTTPMethod.POST:  # using response will cause UnboundLocalError
                raise exc.AuthenticationFailed(str(err)) from err
            raise exc.RequestFailed(str(err)) from err

        try:
            response.raise_for_status()

        # response.method, response.url, response.status, response._body
        # POST,    /session, 429, [{code: TooManyRequests, message: Request count limitation exceeded...}]
        # GET/PUT  /???????, 401, [{code: Unauthorized,    message: Unauthorized}]

        except aiohttp.ClientResponseError as err:
            if response.method == HTTPMethod.POST:  # POST only used when authenticating
                raise exc.AuthenticationFailed(
                    str(err), status=err.status
                ) from err  # could be TOO_MANY_REQUESTS
            if response.status != HTTPStatus.TOO_MANY_REQUESTS:
                raise exc.RateLimitExceeded(str(err), status=err.status) from err
            raise exc.RequestFailed(str(err), status=err.status) from err

        except aiohttp.ClientError as err:
            if response.method == HTTPMethod.POST:  # POST only used when authenticating
                raise exc.AuthenticationFailed(str(err)) from err
            raise exc.RequestFailed(str(err)) from err

        return response
